@@ -259,7 +259,45 @@ int select_previous(Undoptr thisrecord)
 }
 
 /*----------------------------------------------------------------------*/
-/* Put element "thiselem" back into object "thisobj"			*/
+/* This is similar to the above routine, but we just return a pointer	*/
+/* to the index list in the uselection record.  This lets the undelete	*/
+/* function restore the original ordering of parts that were deleted.	*/
+/*----------------------------------------------------------------------*/
+
+short *recover_selectlist(Undoptr thisrecord)
+{
+   Undoptr chkrecord;
+   uselection *srec;
+
+   for (chkrecord = thisrecord->next; chkrecord != NULL; chkrecord = chkrecord->next) {
+
+      /* Selections may cross page changes, but only if in the same series */
+      if ((chkrecord->thisinst != thisrecord->thisinst) &&
+		(chkrecord->idx != thisrecord->idx))
+	 break;
+
+      switch (chkrecord->type) {
+	 case XCF_Delete: case XCF_Pop: case XCF_Push:
+	    /* Delete/Push/Pop records imply a canceled selection */
+	    return NULL;
+	 case XCF_Copy:
+	    /* Copy is the same as Select, but the copied objects are	*/
+	    /* always at the end of the element list.  Returning NULL	*/
+	    /* is the same as declaring that elements should be		*/
+	    /* appended to the end of the object's element list.	*/
+	    return NULL;
+	 case XCF_Select:
+	    srec = (uselection *)chkrecord->undodata;
+	    return srec->idx;
+      }
+   }
+   return NULL;
+}
+
+/*----------------------------------------------------------------------*/
+/* Put element "thiselem" back into object "thisobj".  This is only	*/
+/* used in the case where the elements were previously at the end of	*/
+/* the object's element list.						*/
 /*----------------------------------------------------------------------*/
 
 void undelete_one_element(objinstptr thisinst, genericptr thiselem)
@@ -511,7 +549,8 @@ short undo_one_action()
 	 thisobj = (objectptr)thisrecord->undodata;
 	 areastruct.selects = thisobj->parts;
 	 areastruct.selectlist = xc_undelete(thisrecord->thisinst,
-			thisobj, (short)thisrecord->idata);
+			thisobj, (short)thisrecord->idata,
+			recover_selectlist(thisrecord));
 	 srec = remember_selection(thisrecord->thisinst, areastruct.selectlist,
 			areastruct.selects);
 	 thisrecord->undodata = (char *)srec;
@@ -540,7 +579,8 @@ short undo_one_action()
 	 /* Put back all the parts */
 	 areastruct.selects = thisobj->parts;
 	 areastruct.selectlist = xc_undelete(thisrecord->thisinst,
-			thisobj, (short)thisrecord->idata);
+			thisobj, (short)thisrecord->idata,
+			recover_selectlist(thisrecord));
 	 srec = remember_selection(thisrecord->thisinst, areastruct.selectlist,
 			areastruct.selects);
 	 thisrecord->undodata = (char *)srec;
@@ -786,7 +826,7 @@ short redo_one_action()
 	 if (thisobj != NULL) {
 	    unselect_all();
 	    snum = thisobj->parts;
-	    slist = xc_undelete(thisrecord->thisinst, thisobj, DRAW);
+	    slist = xc_undelete(thisrecord->thisinst, thisobj, DRAW, NULL);
 	    thisrecord->undodata = (char *)remember_selection(thisrecord->thisinst,
 			slist, snum);
 	    free(slist);
@@ -798,7 +838,8 @@ short redo_one_action()
 	 if (thisobj != NULL) {
 	    unselect_all();
 	    areastruct.selects = thisobj->parts;
-	    areastruct.selectlist = xc_undelete(thisrecord->thisinst, thisobj, DRAW);
+	    areastruct.selectlist = xc_undelete(thisrecord->thisinst, thisobj, DRAW,
+			NULL);
 	    thisrecord->undodata = (char *)remember_selection(thisrecord->thisinst,
 			areastruct.selectlist, areastruct.selects);
 	    draw_all_selected();
